﻿// ScatterPlotVisuals.cs by Charles Petzold, December 2008
using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;

namespace DataLibrary
{
    public class ScatterPlotVisuals : FrameworkElement
    {
        VisualCollection visualChildren;

        public static readonly DependencyProperty ItemsSourceProperty =
            DependencyProperty.Register("ItemsSource",
                typeof(ObservableNotifiableCollection<DataPoint>),
                typeof(ScatterPlotVisuals),
                new PropertyMetadata(OnItemsSourceChanged));

        public static readonly DependencyProperty BrushesProperty =
            DependencyProperty.Register("Brushes",
                typeof(Brush[]),
                typeof(ScatterPlotVisuals),
                new FrameworkPropertyMetadata(null, 
                        FrameworkPropertyMetadataOptions.AffectsRender));

        public static readonly DependencyProperty BackgroundProperty =
            Panel.BackgroundProperty.AddOwner(typeof(ScatterPlotRender));

        public ScatterPlotVisuals()
        {
            visualChildren = new VisualCollection(this);
            ToolTip = "";
        }

        public ObservableNotifiableCollection<DataPoint> ItemsSource
        {
            set { SetValue(ItemsSourceProperty, value); }
            get { return (ObservableNotifiableCollection<DataPoint>)GetValue(ItemsSourceProperty); }
        }

        public Brush[] Brushes
        {
            set { SetValue(BrushesProperty, value); }
            get { return (Brush[])GetValue(BrushesProperty); }
        }

        public Brush Background
        {
            set { SetValue(BackgroundProperty, value); }
            get { return (Brush)GetValue(BackgroundProperty); }
        }

        static void OnItemsSourceChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
            (obj as ScatterPlotVisuals).OnItemsSourceChanged(args);
        }

        void OnItemsSourceChanged(DependencyPropertyChangedEventArgs args)
        {
            visualChildren.Clear();

            if (args.OldValue != null)
            {
                ObservableNotifiableCollection<DataPoint> coll = args.OldValue as ObservableNotifiableCollection<DataPoint>;
                coll.CollectionCleared -= OnCollectionCleared;
                coll.CollectionChanged -= OnCollectionChanged;
                coll.ItemPropertyChanged -= OnItemPropertyChanged;
            }

            if (args.NewValue != null)
            {
                ObservableNotifiableCollection<DataPoint> coll = args.NewValue as ObservableNotifiableCollection<DataPoint>;
                coll.CollectionCleared += OnCollectionCleared;
                coll.CollectionChanged += OnCollectionChanged;
                coll.ItemPropertyChanged += OnItemPropertyChanged;

                CreateVisualChildren(coll);
            }
        }

        void OnCollectionCleared(object sender, EventArgs args)
        {
            RemoveVisualChildren(visualChildren);
        }

        void OnCollectionChanged(object sender, NotifyCollectionChangedEventArgs args)
        {
            if (args.OldItems != null)
                RemoveVisualChildren(args.OldItems);

            if (args.NewItems != null)
                CreateVisualChildren(args.NewItems);
        }

        void OnItemPropertyChanged(object sender, ItemPropertyChangedEventArgs args)
        {
            DataPoint dataPoint = args.Item as DataPoint;

            foreach (Visual child in visualChildren)
            {
                DrawingVisualPlus drawingVisual = child as DrawingVisualPlus;

                if (dataPoint == drawingVisual.DataPoint)
                {
                    // Assume only VariableX or VariableY are changing
                    TranslateTransform xform = drawingVisual.Transform as TranslateTransform;

                    if (args.PropertyName == "VariableX")
                        xform.X = RenderSize.Width * dataPoint.VariableX;

                    else if (args.PropertyName == "VariableY")
                        xform.Y = RenderSize.Height * dataPoint.VariableY;
                }
            }
        }

        void CreateVisualChildren(ICollection coll)
        {
            foreach (object obj in coll)
            {
                DataPoint dataPoint = obj as DataPoint;

                DrawingVisualPlus drawingVisual = new DrawingVisualPlus();
                drawingVisual.DataPoint = dataPoint;
                DrawingContext dc = drawingVisual.RenderOpen();

                dc.DrawEllipse(Brushes[dataPoint.Type], null,
                    new Point(0, 0), 1, 1);

                drawingVisual.Transform = new TranslateTransform(RenderSize.Width * dataPoint.VariableX,
                                                                 RenderSize.Height * dataPoint.VariableY);

                dc.Close();
                visualChildren.Add(drawingVisual);
            }
        }

        void RemoveVisualChildren(ICollection coll)
        {
            foreach (object obj in coll)
            {
                DataPoint dataPoint = obj as DataPoint;
                List<DrawingVisualPlus> removeList = new List<DrawingVisualPlus>();

                foreach (Visual child in visualChildren)
                {
                    DrawingVisualPlus drawingVisual = child as DrawingVisualPlus;
                    if (drawingVisual.DataPoint == dataPoint)
                    {
                        removeList.Add(drawingVisual);
                        break;
                    }
                }
                foreach (DrawingVisualPlus drawingVisual in removeList)
                    visualChildren.Remove(drawingVisual);
            }
        }

        protected override void OnRenderSizeChanged(SizeChangedInfo sizeInfo)
        {
            foreach (Visual child in visualChildren)
            {
                DrawingVisualPlus drawingVisual = child as DrawingVisualPlus;
                TranslateTransform xform = drawingVisual.Transform as TranslateTransform;

                if (sizeInfo.WidthChanged)
                    xform.X = sizeInfo.NewSize.Width * drawingVisual.DataPoint.VariableX;

                if (sizeInfo.HeightChanged)
                    xform.Y = sizeInfo.NewSize.Height * drawingVisual.DataPoint.VariableY;
            }
            base.OnRenderSizeChanged(sizeInfo);
        }

        protected override int VisualChildrenCount
        {
            get
            {
                return visualChildren.Count;
            }
        }

        protected override Visual GetVisualChild(int index)
        {
            if (index < 0 || index >= visualChildren.Count)
                throw new ArgumentOutOfRangeException("index");

            return visualChildren[index];
        }

        protected override void OnRender(DrawingContext dc)
        {
            dc.DrawRectangle(Background, null, new Rect(RenderSize));
        }

        protected override void OnToolTipOpening(ToolTipEventArgs e)
        {
            HitTestResult result = VisualTreeHelper.HitTest(this, Mouse.GetPosition(this));

            if (result.VisualHit is DrawingVisualPlus)
            {
                DrawingVisualPlus drawingVisual = result.VisualHit as DrawingVisualPlus;
                DataPoint dataPoint = drawingVisual.DataPoint;
                ToolTip = String.Format("{0}, X={1}, Y={2}", dataPoint.ID, dataPoint.VariableX, dataPoint.VariableY);
            }
            base.OnToolTipOpening(e);
        }

        protected override void OnToolTipClosing(ToolTipEventArgs e)
        {
            ToolTip = "";
            base.OnToolTipClosing(e);
        }

        class DrawingVisualPlus : DrawingVisual
        {
            public DataPoint DataPoint { get; set; }
        }
    }
}
